# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""High level analog tiles (inference)."""

from copy import deepcopy
from typing import List, Optional, Union, TYPE_CHECKING

from torch import device as torch_device
from torch import ones, zeros, Tensor
from torch.autograd import no_grad

from aihwkit.exceptions import CudaError
from aihwkit.simulator.rpu_base import cuda
from aihwkit.simulator.tiles.analog import AnalogTile

if TYPE_CHECKING:
    from aihwkit.simulator.configs import InferenceRPUConfig
    from aihwkit.simulator.tiles import BaseTile

# pylint: disable=too-many-instance-attributes


class InferenceTile(AnalogTile):
    """Tile used for analog inference and hardware-aware training for inference.

    Args:
        out_size: output size
        in_size: input size
        rpu_config: resistive processing unit configuration.
        bias: whether to add a bias column to the tile.
        in_trans: Whether to assume an transposed input (batch first)
        out_trans: Whether to assume an transposed output (batch first)
        shared_weights: Whether to keep the weight in torch's memory space
    """

    def __init__(
            self,
            out_size: int,
            in_size: int,
            rpu_config: Optional['InferenceRPUConfig'] = None,
            bias: bool = False,
            in_trans: bool = False,
            out_trans: bool = False,
            shared_weights: bool = True,
    ):

        if not rpu_config:
            # Import `InferenceRPUConfig` dynamically to avoid import cycles.
            # pylint: disable=import-outside-toplevel
            from aihwkit.simulator.configs import InferenceRPUConfig
            rpu_config = InferenceRPUConfig()

        # Noise model.
        self.noise_model = deepcopy(rpu_config.noise_model)

        # Drift compensation.
        self.drift_compensation = None
        if rpu_config.drift_compensation:
            self.drift_compensation = deepcopy(rpu_config.drift_compensation)
        self.drift_baseline = None
        self.drift_readout_tensor = None  # type: Optional[Tensor]
        self.alpha = ones((1,))

        # Helpers.
        self.reference_combined_weights = None  # type: Optional[Tensor]
        self.programmed_weights = None  # type: Optional[Tensor]
        self.nu_drift_list = None  # type: Optional[List[Tensor]]

        super().__init__(out_size, in_size, rpu_config, bias, in_trans, out_trans)

        if shared_weights:
            self.shared_weights = zeros(out_size, in_size + int(bias),
                                        requires_grad=True)  # type: Tensor
            self.ensure_shared_weights()

    @no_grad()
    def _forward_drift_readout_tensor(self) -> Optional[Tensor]:
        """Perform a forward pass using the drift read-out tensor."""
        if self.drift_compensation is None:
            return None

        if self.drift_readout_tensor is None:
            self.drift_readout_tensor = self.drift_compensation.get_readout_tensor(
                self.tile.get_x_size()).detach().to(self.device)
            if self.in_trans:
                self.drift_readout_tensor = self.drift_readout_tensor.tranpose(0, 1).clone()

        # We need to take the bias as a common column here, also we do
        # not want to use indexed.
        return self.tile.forward(self.drift_readout_tensor, False,
                                 self.in_trans, self.out_trans, True)

    @no_grad()
    def program_weights(self, from_reference: bool = True) -> None:
        """Apply weights noise to the current tile weights and saves these for
        repeated drift experiments.

        This method also establishes the drift coefficients for each
        conductance slice.

        Args:
            from_reference: Whether to use weights from reference
        """
        if not from_reference or self.reference_combined_weights is None:
            self.reference_combined_weights = Tensor(self.tile.get_weights())

        self.programmed_weights, self.nu_drift_list = self.noise_model.apply_programming_noise(
            self.reference_combined_weights)

        if self.drift_compensation is not None:
            self.tile.set_weights(self.programmed_weights.numpy())
            forward_output = self._forward_drift_readout_tensor()

            self.drift_baseline = self.drift_compensation.init_baseline(forward_output)

    @no_grad()
    def drift_weights(
            self,
            t_inference: float = 0.0
    ) -> None:
        """Programs and drifts the current reference weights.

        The current weight reference is either the current weights or
        the ones at the time when :meth:`initialize_drift_reference`
        was called, which then would overwrite the current weights
        with the drifted ones.

        Args:
            t_inference: Time (in sec) of assumed inference
                time. Programming ends at t=0s.  The rest is waiting time,
                where the devices might drift and accumulate noise. See
                noise model used for details.
        """
        # pylint: disable=arguments-differ
        if self.programmed_weights is None:
            self.program_weights()

        drifted_weights = self.noise_model.apply_drift_noise(
            self.programmed_weights, self.nu_drift_list, t_inference)
        self.tile.set_weights(drifted_weights.detach().cpu().numpy())

        if self.drift_compensation is not None:
            forward_output = self._forward_drift_readout_tensor()
            self.alpha = self.drift_compensation.apply(forward_output,
                                                       self.drift_baseline).to(self.device)

    def forward(self, x_input: Tensor, is_test: bool = False) -> Tensor:
        """Forward pass with drift compensation.

        Note:
            The drift compensation scale will only be applied during
            testing, ie if ``is_test=True``.
        """
        # Import `aihwkit.simulator.configs` items dynamically to avoid import cycles.
        # pylint: disable=import-outside-toplevel
        from aihwkit.simulator.configs.helpers import parameters_to_bindings
        from aihwkit.simulator.configs.utils import WeightModifierType

        if not is_test and (self.rpu_config.modifier.type != WeightModifierType.COPY or
                            self.rpu_config.modifier.pdrop > 0.0):
            weight_modify_params = parameters_to_bindings(self.rpu_config.modifier)
            self.tile.modify_weights(weight_modify_params)

        if not is_test or self.drift_compensation is None:
            return super().forward(x_input, is_test)

        # only do drift compensation in eval mode
        return super().forward(x_input, True)*self.alpha

    @no_grad()
    def post_update_step(self) -> None:
        """Operators that need to be called once per mini-batch."""
        # Import `aihwkit.simulator.configs` items dynamically to avoid import cycles.
        # pylint: disable=import-outside-toplevel
        from aihwkit.simulator.configs.helpers import parameters_to_bindings
        from aihwkit.simulator.configs.utils import WeightClipType

        super().post_update_step()

        # TODO: make this a little nicer. Now each time bindings are generated.
        if self.rpu_config.clip.type != WeightClipType.NONE:
            weight_clip_params = parameters_to_bindings(self.rpu_config.clip)
            self.tile.clip_weights(weight_clip_params)

    def cuda(
            self,
            device: Optional[Union[torch_device, str, int]] = None
    ) -> 'BaseTile':
        """Return a copy of this tile in CUDA memory.

        Args:
            device: CUDA device

        Returns:
            Self with the underlying C++ tile moved to CUDA memory.

        Raises:
            CudaError: if the library has not been compiled with CUDA.
        """
        super().cuda(device)

        self.alpha = self.alpha.cuda(device)
        self.shared_weights.data = zeros(self.tile.get_x_size(),
                                         self.tile.get_d_size(),
                                         requires_grad=True).cuda(device)
        self.ensure_shared_weights()

        return self


class CudaInferenceTile(InferenceTile):
    """Analog inference tile (CUDA).

    Analog inference tile that uses GPU for its operation. The instantiation is based on
    an existing non-cuda tile: all the source attributes are copied except
    for the simulator tile, which is recreated using a GPU tile.

    Caution:
        Deprecated. Use ``InferenceTile(..).cuda()`` instead.

    Args:
        source_tile: tile to be used as the source of this tile
    """

    def __init__(self, source_tile: AnalogTile):
        if not cuda.is_compiled():
            raise CudaError('aihwkit has not been compiled with CUDA support')

        # Create a new instance of the rpu config.
        new_rpu_config = deepcopy(source_tile.rpu_config)

        # Create the tile, replacing the simulator tile.
        super().__init__(source_tile.out_size, source_tile.in_size, new_rpu_config,
                         source_tile.bias, source_tile.in_trans, source_tile.out_trans,
                         shared_weights=True)

        self.cuda(self.device)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Runner that executes Experiments locally."""

from typing import Dict, List, Optional

from torch import device as torch_device
from torchvision.datasets import FashionMNIST, SVHN

from aihwkit.experiments.experiments.base import Signals
from aihwkit.experiments.runners.base import Runner
from aihwkit.experiments.experiments.training import BasicTraining
from aihwkit.experiments.runners.metrics import LocalMetric


class LocalRunner(Runner):
    """Runner that executes Experiments locally.

    Class that allows executing Experiments locally.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self,
                 device: Optional[torch_device] = None):
        """Create a new ``LocalRunner``.

        Args:
            device: the device where the model will be moved to.
        """
        self.device = device

    def run(  # type: ignore[override]
            self,
            experiment: BasicTraining,
            max_elements_train: int = 0,
            dataset_root: str = '/tmp/datasets',
            stdout: bool = True
    ) -> List[Dict]:
        """Run a single Experiment.

        Executes an experiment locally, in the device specified by
        ``self.device``, optionally printing information to stdout.

        Note:
            If using a dataset different than ``FashionMNIST`` or ``SVHN``,
            the runner assumes that the files for the dataset are downloaded at
            ``dataset_root``. For those two datasets, the downloading will
            take place automatically if the files are not present.

        Args:
            experiment: the experiment to be executed.
            max_elements_train: limit on the amount of samples to use from
                the dataset. If ``0``, no limit is applied.
            dataset_root: path for the dataset files.
            stdout: enable printing to stdout during the execution of the
                experiment.

        Returns:
            A list of dictionaries with information about each epoch.
        """
        # pylint: disable=arguments-differ

        # Setup the metric helper for the experiment.
        metric = LocalMetric(stdout=stdout)
        experiment.clear_hooks()
        experiment.add_hook(Signals.EPOCH_START, metric.receive_epoch_start)
        experiment.add_hook(Signals.EPOCH_END, metric.receive_epoch_end)
        experiment.add_hook(Signals.TRAIN_EPOCH_END, metric.receive_train_epoch_end)
        experiment.add_hook(Signals.TRAIN_EPOCH_BATCH_END, metric.receive_train_epoch_batch_end)
        experiment.add_hook(Signals.VALIDATION_EPOCH_BATCH_END,
                            metric.receive_validation_epoch_batch_end)
        experiment.add_hook(Signals.VALIDATION_EPOCH_END, metric.receive_validation_epoch_end)

        # Download the dataset if needed.
        if experiment.dataset == FashionMNIST:
            _ = experiment.dataset(dataset_root, download=True)
        elif experiment.dataset == SVHN:
            _ = experiment.dataset(dataset_root, download=True, split='train')
            _ = experiment.dataset(dataset_root, download=True, split='test')

        # Move the model to the device.
        if self.device:
            if self.device.type == 'cuda':
                experiment.model = experiment.model.to(self.device)

        # Build the objects needed for training.
        training_loader, validation_loader = experiment.get_data_loaders(
            experiment.dataset, experiment.batch_size,
            max_elements_train=max_elements_train,
            dataset_root=dataset_root
        )
        optimizer = experiment.get_optimizer(experiment.learning_rate, experiment.model)

        experiment.results = experiment.train(training_loader,
                                              validation_loader,
                                              experiment.model,
                                              optimizer,
                                              experiment.loss_function(),
                                              experiment.epochs,
                                              self.device)
        return experiment.results



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for model conversions."""

from torch import Tensor, manual_seed
from torch.nn import Sequential, Linear
from torch.nn.functional import mse_loss
from torch.optim import SGD
from torchvision.models import resnet18, alexnet
from parameterized import parameterized_class

from aihwkit.simulator.configs.configs import FloatingPointRPUConfig
from aihwkit.nn import AnalogSequential, AnalogLinear, AnalogConv2d
from aihwkit.nn.conversion import convert_to_analog

from .helpers.testcases import ParametrizedTestCase


@parameterized_class([
    {'use_cuda': False, 'name': 'Cuda'},
    {'use_cuda': True, 'name': 'Cpu'},
], class_name_func=lambda cls, _, params_dict: '%s%s' % (cls.__name__, params_dict['name']))
class ConversionLayerTest(ParametrizedTestCase):
    """Linear layer abstractions tests."""

    @staticmethod
    def train_model_torch(model, loss_func, x_b, y_b):
        """Train the model with torch SGD."""
        opt = SGD(model.parameters(), lr=0.5)
        epochs = 100
        for _ in range(epochs):
            opt.zero_grad()
            pred = model(x_b)
            loss = loss_func(pred, y_b)

            loss.backward()
            opt.step()

    def test_conversion_linear_sequential(self):
        """Test converting sequential and linear."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)
        model = Sequential(
            Linear(4, 3),
            Linear(3, 3),
            Sequential(
                Linear(3, 1),
                Linear(1, 1)
            )
        )
        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            model = model.cuda()

        self.train_model_torch(model, loss_func, x_b, y_b)
        digital_loss = loss_func(model(x_b), y_b)
        analog_model = convert_to_analog(model, FloatingPointRPUConfig())
        self.assertEqual(analog_model[0].__class__, AnalogLinear)
        self.assertEqual(analog_model.__class__, AnalogSequential)
        self.assertTensorAlmostEqual(loss_func(analog_model(x_b), y_b), digital_loss)

    def test_conversion_torchvision_resnet(self):
        """Test converting resnet model from torchvision."""
        model = resnet18()
        if self.use_cuda:
            model = model.cuda()

        analog_model = convert_to_analog(model, FloatingPointRPUConfig())
        self.assertEqual(analog_model.conv1.__class__, AnalogConv2d)
        self.assertEqual(analog_model.layer1.__class__, AnalogSequential)
        self.assertEqual(analog_model.layer1[0].conv1.__class__, AnalogConv2d)

    def test_conversion_torchvision_alexnet(self):
        """Test converting resnet model from torchvision"""
        model = alexnet()
        if self.use_cuda:
            model = model.cuda()

        analog_model = convert_to_analog(model, FloatingPointRPUConfig())
        self.assertEqual(analog_model.features[0].__class__, AnalogConv2d)
        self.assertEqual(analog_model.classifier[6].__class__, AnalogLinear)
        self.assertEqual(analog_model.features.__class__, AnalogSequential)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Analog-aware inference optimizer."""

from types import new_class
from typing import Any, Callable, Dict, Optional, Type

from torch.optim import Optimizer, SGD

from aihwkit.optim.context import AnalogContext


class AnalogOptimizerMixin:
    """Mixin for analog optimizers.

    This class contains the methods needed for enabling analog in an existing
    ``Optimizer``. It is designed to be used as a mixin in conjunction with an
    ``AnalogOptimizer`` or torch ``Optimizer``.
    """

    def regroup_param_groups(self, *_: Any) -> None:
        """Reorganize the parameter groups, isolating analog layers.

        Update the `param_groups` of the optimizer, moving the parameters for
        each analog layer to a new single group.
        """
        # Create the new param groups.
        analog_param_groups = []
        rm_group_lst = []
        for group in self.param_groups:  # type: ignore[has-type]
            rm_lst = []
            for param in group['params']:
                if isinstance(param, AnalogContext):
                    param.analog_tile.set_learning_rate(
                        self.defaults['lr'])  # type: ignore[attr-defined]
                    analog_param_groups.append({
                        'params': [param],
                    })
                    rm_lst.append(id(param))

            group['params'] = [p for p in group['params'] if id(p) not in rm_lst]

            if len(group['params']) == 0:
                rm_group_lst.append(id(group))

        self.param_groups = [g for g in self.param_groups  # type: ignore[has-type]
                             if id(g) not in rm_group_lst]

        # Add analog groups.
        for group in analog_param_groups:
            self.add_param_group(group)  # type: ignore[attr-defined]

    def step(self, closure: Optional[Callable] = None) -> Optional[float]:
        """Perform an analog-aware single optimization step.

        If a group containing analog parameters is detected, the optimization
        step calls the related RPU controller. For regular parameter groups,
        the optimization step has the same behaviour as ``torch.optim.SGD``.

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.

        Returns:
            The loss, if ``closure`` has been passed as a parameter.
        """
        # pylint: disable=too-many-branches
        # Update non-analog parameters using the given optimizer
        ret = super().step(closure)  # type: ignore[misc]

        # Update analog parameters
        for group in self.param_groups:
            learning_rate = group.get('lr')

            # Use analog_tile object.
            for param in group['params']:
                if isinstance(param, AnalogContext):
                    # Handle internal analog update.
                    analog_ctx = param
                    analog_tile = analog_ctx.analog_tile

                    if analog_ctx.use_torch_update:
                        # In this case a separate weight parameter exists: do nothing.
                        continue

                    # Update learning rate.
                    if learning_rate is not None:
                        analog_tile.set_learning_rate(learning_rate)

                    # Call `update` in the tile.
                    if not analog_ctx.has_gradient():
                        # Forward never used.
                        continue

                    if analog_ctx.use_indexed:
                        for x_input, d_input in zip(analog_ctx.analog_input,
                                                    analog_ctx.analog_grad_output):
                            analog_tile.update_indexed(x_input, d_input)
                    else:
                        for x_input, d_input in zip(analog_ctx.analog_input,
                                                    analog_ctx.analog_grad_output):
                            analog_tile.update(x_input, d_input)

                    analog_ctx.reset()
        # Apply post-update step operations (diffuse, decay, etc).
        # (only here because of unknown params order and shared weights)
        for group in self.param_groups:
            for param in group['params']:
                if isinstance(param, AnalogContext):
                    param.analog_tile.post_update_step()

        return ret

    def set_learning_rate(self, learning_rate: float = 0.1) -> None:
        """Update the learning rate to a new value.

        Update the learning rate of the optimizer, propagating the changes
        to the analog tiles accordingly.

        Args:
            learning_rate: learning rate for the optimizer.
        """
        for param_group in self.param_groups:
            param_group['lr'] = learning_rate
            for param in param_group['params']:
                if isinstance(param, AnalogContext):
                    # Update learning rate on the tile
                    param.analog_tile.set_learning_rate(learning_rate)


class AnalogOptimizer(AnalogOptimizerMixin, Optimizer):
    """Generic optimizer that wraps an existing ``Optimizer`` for analog inference.

    This class wraps an existing ``Optimizer``, customizing the optimization
    step for triggering the analog update needed for analog tiles. All other
    (digital) parameters are governed by the given torch optimizer. In case of
    hardware-aware training (``InferenceTile``) the tile weight update is also
    governed by the given optimizer, otherwise it is using the internal analog
    update as defined in the ``rpu_config``.

    The ``AnalogOptimizer`` constructor expects the wrapped optimizer class as
    the first parameter, followed by any arguments required by the wrapped
    optimizer.

    Note:
        The instances returned are of a *new* type that is a subclass of:

        * the wrapped ``Optimizer`` (allowing access to all their methods and
          attributes).
        * this ``AnalogOptimizer``.

    Example:
        The following block illustrate how to create an optimizer that wraps
        standard SGD:

        >>> from torch.optim import SGD
        >>> from torch.nn import Linear
        >>> from aihwkit.simulator.configs.configs import InferenceRPUConfig
        >>> from aihwkit.optim import AnalogOptimizer
        >>> model = AnalogLinear(3, 4, rpu_config=InferenceRPUConfig)
        >>> optimizer = AnalogOptimizer(SGD, model.parameters(), lr=0.02)
    """

    SUBCLASSES = {}  # type: Dict[str, Type]
    """Registry of the created subclasses."""

    def __new__(
            cls,
            optimizer_cls: Type,
            *_: Any,
            **__: Any
    ) -> 'AnalogOptimizer':
        subclass_name = '{}{}'.format(cls.__name__, optimizer_cls.__name__)

        # Retrieve or create a new subclass, that inherits both from
        # `AnalogOptimizer` and for the specific torch optimizer
        # (`optimizer_cls`).
        if subclass_name not in cls.SUBCLASSES:
            cls.SUBCLASSES[subclass_name] = new_class(subclass_name, (cls, optimizer_cls), {})

        return super().__new__(cls.SUBCLASSES[subclass_name])

    def __init__(
            self,
            optimizer_cls: Type,  # pylint: disable=unused-argument
            *args: Any,
            **kwargs: Any
    ):
        super().__init__(*args, **kwargs)


class AnalogSGD(AnalogOptimizerMixin, SGD):
    """Implements analog-aware stochastic gradient descent."""



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Base class for an Experiment."""

from enum import Enum
from typing import Any, Callable, Dict, Optional


class Signals(Enum):
    """Signals emitted by an Experiment."""

    EXPERIMENT_START = 1
    EXPERIMENT_END = 2
    EPOCH_START = 10
    EPOCH_END = 11
    TRAIN_EPOCH_START = 20
    TRAIN_EPOCH_END = 21
    TRAIN_EPOCH_BATCH_START = 22
    TRAIN_EPOCH_BATCH_END = 23
    VALIDATION_EPOCH_START = 30
    VALIDATION_EPOCH_END = 31
    VALIDATION_EPOCH_BATCH_START = 32
    VALIDATION_EPOCH_BATCH_END = 33


class Experiment:
    """Base class for an Experiment.

    This class is used as the base class for more specific experiments. The
    experiments use ``hooks`` for reporting the different status changes to the
    ``Metrics`` during the execution of the experiment.
    """

    def __init__(self) -> None:
        self.hooks: Dict = {
            Signals.EXPERIMENT_START: [],
            Signals.EXPERIMENT_END: [],
            Signals.EPOCH_START: [],
            Signals.EPOCH_END: [],
            Signals.TRAIN_EPOCH_START: [],
            Signals.TRAIN_EPOCH_END: [],
            Signals.TRAIN_EPOCH_BATCH_START: [],
            Signals.TRAIN_EPOCH_BATCH_END: [],
            Signals.VALIDATION_EPOCH_START: [],
            Signals.VALIDATION_EPOCH_END: [],
            Signals.VALIDATION_EPOCH_BATCH_START: [],
            Signals.VALIDATION_EPOCH_BATCH_END: []
        }

        self.results: Optional[Any] = None

    def add_hook(self, key: Signals, hook: Callable) -> None:
        """Register a hook for the experiment.

        Register a new hook for a particular signal. During the execution of
        the experiment, the ``hook`` function will be called.

        Args:
            key: signal which the hook will be registered to.
            hook: a function that will be called when the signal is emitted.
        """
        self.hooks[key].append(hook)

    def clear_hooks(self) -> None:
        """Remove all the hooks from the experiment."""
        for key in self.hooks:
            self.hooks[key] = []

    def _call_hook(self, key: Signals, *args: Any, **kwargs: Any) -> Dict:
        """Invoke the hooks for a specific key."""
        ret = {}
        for hook in self.hooks[key]:
            hook_ret = hook(*args, **kwargs)
            if isinstance(hook_ret, dict):
                ret.update(hook_ret)

        return ret



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Base class for analog Modules."""

from typing import Any, Dict, List, Optional, Tuple, Union, TYPE_CHECKING

from torch import Tensor
from torch.nn import Module, Parameter

from aihwkit.exceptions import ModuleError
from aihwkit.simulator.configs import (
    FloatingPointRPUConfig, InferenceRPUConfig, SingleRPUConfig,
    UnitCellRPUConfig
)
from aihwkit.simulator.tiles import InferenceTile
if TYPE_CHECKING:
    from aihwkit.simulator.tiles import BaseTile

RPUConfigAlias = Union[FloatingPointRPUConfig, SingleRPUConfig,
                       UnitCellRPUConfig, InferenceRPUConfig]


class AnalogModuleBase(Module):
    """Base class for analog Modules.

    Base ``Module`` for analog layers that use analog tiles. When subclassing,
    please note:

    * the ``_setup_tile()`` method is expected to be called by the subclass
      constructor, and it does not only create a tile, but also sets some
      instance attributes that are needed by the analog features (optimizer
      and others).
    * the ``weight`` and ``bias`` Parameters are not guaranteed to be in
      sync with the tile weights and biases during the lifetime of the instance,
      for performance reasons. The canonical way of reading and writing
      weights is via the ``set_weights()`` and ``get_weights()`` as opposed
      to using the attributes directly.
    * the ``BaseTile`` subclass that is created is retrieved from the
      ``rpu_config.tile_class`` attribute.
    """
    # pylint: disable=abstract-method
    _analog_tile_counter: int = 0

    def register_analog_tile(self, tile: 'BaseTile') -> None:
        """Register the analog context of the tile.

        Note:
            Needs to be called at the end init to register the tile
            for the analog optimizers.

        Args:
            tile: tile to register
        """
        self.register_parameter('analog_ctx_' + str(self._analog_tile_counter),
                                tile.get_analog_ctx())
        if tile.shared_weights is not None:
            if not isinstance(tile.shared_weights, Parameter):
                tile.shared_weights = Parameter(tile.shared_weights)
            self.register_parameter('analog_shared_weights_' + str(self._analog_tile_counter),
                                    tile.shared_weights)

        self._analog_tile_counter += 1

    def unregister_parameter(self, param_name: str) -> None:
        """Unregister module parameter from parameters.

        Raises:
            ModuleError: In case parameter is not found
        """
        param = getattr(self, param_name, None)
        if not isinstance(param, Parameter):
            raise ModuleError(f"Cannot find parameter {param_name} to unregister")
        param_data = param.detach().clone()
        delattr(self, param_name)
        setattr(self, param_name, param_data)

    def _setup_tile(
            self,
            in_features: int,
            out_features: int,
            bias: bool,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: float = 0.0
    ) -> 'BaseTile':
        """Create an analog tile and setup this layer for using it.

        Create an analog tile to be used for the basis of this layer operations,
        and setup additional attributes of this instance that are needed for
        using the analog tile.

        If ``weight_scaling_omega`` is larger than 0, the weights are set in a
        scaled manner (assuming a digital output scale). See
        :meth:`~aihwkit.simulator.tiles.base.BaseTile.set_weights_scaled`
        for details.

        Note:
            This method also sets the following attributes, which are assumed
            to be set by the rest of the methods:
            * ``self.use_bias``
            * ``self.realistic_read_write``
            * ``self.weight_scaling_omega``
            * ``self.in_features``
            * ``self.out_features``

        Args:
            in_features: input vector size (number of columns).
            out_features: output vector size (number of rows).
            rpu_config: resistive processing unit configuration.
            bias: whether to use a bias row on the analog tile or not.
            realistic_read_write: whether to enable realistic read/write
                for setting initial weights and read out of weights.
            weight_scaling_omega: the weight value where the max
                weight will be scaled to. If zero, no weight scaling will
                be performed

        Returns:
            An analog tile with the requested parameters.
        """
        # pylint: disable=attribute-defined-outside-init, protected-access
        # Default to constant step device if not provided.
        if not rpu_config:
            rpu_config = SingleRPUConfig()

        # Setup the analog-related attributes of this instance.
        self.use_bias = bias
        self.realistic_read_write = realistic_read_write
        self.weight_scaling_omega = weight_scaling_omega
        self.in_features = in_features
        self.out_features = out_features

        # Create the tile.
        return rpu_config.tile_class(out_features, in_features, rpu_config, bias=bias)

    def set_weights(
            self,
            weight: Tensor,
            bias: Optional[Tensor] = None,
            force_exact: bool = False
    ) -> None:
        """Set the weight (and bias) with given Tensors.

        This uses an realistic write if the property ``realistic_read_write``
        of the layer is set, unless it is overwritten by ``force_exact``. It
        uses a scaled write if ``weight_scaling_omega`` is positive (see
        :meth:`~aihwkit.simulator.tiles.base.BaseTile.set_weights_scaled`).

        Note:
            This is the recommended way for setting the weight/bias matrix of
            the analog tile, as it will correctly store the weights into the
            internal memory. Directly writing to ``self.weight`` and
            ``self.bias`` might yield wrong results as they are not always in
            sync with the analog tile Parameters, for performance reasons.

        Args:
            weight: weight matrix
            bias: bias vector
            force_exact: forces an exact write to the analog tiles
        """
        shape = [self.out_features, self.in_features]
        weight = weight.clone().reshape(shape)

        realistic = self.realistic_read_write and not force_exact

        if self.weight_scaling_omega > 0.0:
            self.analog_tile.set_weights_scaled(weight, bias, realistic=realistic,
                                                omega=self.weight_scaling_omega)
        else:
            self.analog_tile.set_weights(weight, bias, realistic=realistic)

        self._sync_weights_from_tile()

    def get_weights(
            self,
            force_exact: bool = False
    ) -> Tuple[Tensor, Optional[Tensor]]:
        """Get the weight (and bias) tensors.

        This uses an realistic read if the property ``realistic_read_write`` of
        the layer is set, unless it is overwritten by ``force_exact``. It
        scales the analog weights by the digital alpha scale if
        ``weight_scaling_omega`` is positive (see
        :meth:`~aihwkit.simulator.tiles.base.BaseTile.get_weights_scaled`).

        Note:
            This is the recommended way for setting the weight/bias matrix from
            the analog tile, as it will correctly fetch the weights from the
            internal memory. Accessing ``self.weight`` and ``self.bias`` might
            yield wrong results as they are not always in sync with the
            analog tile library, for performance reasons.

        Args:
            force_exact: forces an exact read to the analog tiles

        Returns:
            tuple: weight matrix, bias vector
        """
        realistic = self.realistic_read_write and not force_exact

        if self.weight_scaling_omega > 0.0:
            return self.analog_tile.get_weights_scaled(realistic=realistic)

        return self.analog_tile.get_weights(realistic=realistic)

    def _sync_weights_from_tile(self) -> None:
        """Update the layer weight and bias from the values on the analog tile.

        Update the ``self.weight`` and ``self.bias`` Parameters with an
        exact copy of the internal analog tile weights.
        """
        tile_weight, tile_bias = self.get_weights(force_exact=True)
        self.weight.data[:] = tile_weight.reshape(self.weight.shape)
        if self.use_bias:
            self.bias.data[:] = tile_bias.reshape(self.bias.shape)  # type: ignore

    def _sync_weights_to_tile(self) -> None:
        """Update the tile values from the layer weights and bias.

        Update the internal tile weights with an exact copy of the values of
        the ``self.weight`` and ``self.bias`` Parameters.
        """
        self.set_weights(self.weight, self.bias, force_exact=True)

    def _load_from_state_dict(
            self,
            state_dict: Dict,
            prefix: str,
            local_metadata: Dict,
            strict: bool,
            missing_keys: List[str],
            unexpected_keys: List[str],
            error_msgs: List[str]) -> None:
        """Copy parameters and buffers from `state_dict` into only this
        module, but not its descendants.

        This method is a specialization of ``Module._load_from_state_dict``
        that takes into account the extra ``analog_tile_state`` key used by
        analog layers.
        """
        key = '{}analog_tile_state'.format(prefix)
        if key in state_dict:
            analog_state = state_dict.pop(key)
            self.analog_tile.__setstate__(analog_state)
        elif strict:
            missing_keys.append(key)

        # update the weight / bias (not saved explicitly)
        self._sync_weights_from_tile()

        super()._load_from_state_dict(
            state_dict, prefix, local_metadata, strict, missing_keys,
            unexpected_keys, error_msgs)

    def state_dict(
            self,
            destination: Any = None,
            prefix: str = '',
            keep_vars: bool = False
    ) -> Dict:
        """Return a dictionary containing a whole state of the module."""
        self._sync_weights_from_tile()
        # TODO: this will also pickle the resistive device. Problematic?  we
        # could also just save hidden_pars and weights. However, in any case the
        # state_dict will not reflect the model.parameters() any more, which
        # might get tricky. In any case, internal hidden weights need to be
        # saved to reconstruct a meaningful analog tile

        analog_state = self.analog_tile.__getstate__()
        current_state = super().state_dict(destination, prefix, keep_vars)
        current_state['{}analog_tile_state'.format(prefix)] = analog_state
        return current_state

    def drift_analog_weights(self, t_inference: float = 0.0) -> None:
        """(Program) and drift the analog weights.

        Args:
            t_inference: assumed time of inference (in sec)

        Raises:
            ModuleError: if the layer is not in evaluation mode.
        """
        if self.training:
            raise ModuleError('drift_analog_weights can only be applied in '
                              'evaluation mode')

        if isinstance(self.analog_tile, InferenceTile):
            self.analog_tile.drift_weights(t_inference)

    def program_analog_weights(self) -> None:
        """Program the analog weights.

        Raises:
            ModuleError: if the layer is not in evaluation mode.
        """
        if self.training:
            raise ModuleError('program_analog_weights can only be applied in '
                              'evaluation mode')

        if isinstance(self.analog_tile, InferenceTile):
            self.analog_tile.program_weights()

    def extra_repr(self) -> str:
        """Set the extra representation of the module.

        Returns:
            A string with the extra representation.
        """
        output = super().extra_repr()
        if self.realistic_read_write:
            output += ', realistic_read_write={}'.format(self.realistic_read_write)
        if self.weight_scaling_omega > 0:
            output += ', weight_scaling_omega={:.3f}'.format(self.weight_scaling_omega)

        return output



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Configurations for resistive processing units."""

from dataclasses import dataclass, field
from typing import ClassVar, Type

from aihwkit.simulator.configs.devices import (
    ConstantStepDevice, FloatingPointDevice, IdealDevice, PulsedDevice,
    UnitCell, DigitalRankUpdateCell
)
from aihwkit.simulator.configs.helpers import (
    _PrintableMixin, tile_parameters_to_bindings
)
from aihwkit.simulator.configs.utils import (
    IOParameters, PulseType, UpdateParameters, WeightClipParameter,
    WeightModifierParameter
)
from aihwkit.simulator.noise_models import (
    BaseDriftCompensation, BaseNoiseModel, GlobalDriftCompensation,
    PCMLikeNoiseModel
)
from aihwkit.simulator.rpu_base import devices

from aihwkit.simulator.tiles import AnalogTile, FloatingPointTile, InferenceTile


@dataclass
class FloatingPointRPUConfig(_PrintableMixin):
    """Configuration for a floating point resistive processing unit."""

    tile_class: ClassVar[Type] = FloatingPointTile
    """Tile class that correspond to this RPUConfig."""

    device: FloatingPointDevice = field(default_factory=FloatingPointDevice)
    """Parameters that modify the behavior of the pulsed device."""


@dataclass
class SingleRPUConfig(_PrintableMixin):
    """Configuration for an analog (pulsed device) resistive processing unit."""

    tile_class: ClassVar[Type] = AnalogTile
    """Tile class that correspond to this RPUConfig."""

    bindings_class: ClassVar[Type] = devices.AnalogTileParameter

    device: PulsedDevice = field(default_factory=ConstantStepDevice)
    """Parameters that modify the behavior of the pulsed device."""

    forward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the forward direction."""

    backward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the backward direction."""

    update: UpdateParameters = field(default_factory=UpdateParameters)
    """Parameter for the update behavior."""

    def as_bindings(self) -> devices.AnalogTileParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return tile_parameters_to_bindings(self)


@dataclass
class UnitCellRPUConfig(_PrintableMixin):
    """Configuration for an analog (unit cell) resistive processing unit."""

    tile_class: ClassVar[Type] = AnalogTile
    """Tile class that correspond to this RPUConfig."""

    bindings_class: ClassVar[Type] = devices.AnalogTileParameter

    device: UnitCell = field(default_factory=UnitCell)
    """Parameters that modify the behavior of the pulsed device."""

    forward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the forward direction."""

    backward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the backward direction."""

    update: UpdateParameters = field(default_factory=UpdateParameters)
    """Parameter for the parallel analog update behavior."""

    def as_bindings(self) -> devices.AnalogTileParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return tile_parameters_to_bindings(self)


@dataclass
class InferenceRPUConfig(_PrintableMixin):
    """Configuration for an analog tile that is used only for inference.

    Training is done in *hardware-aware* manner, thus using only the
    non-idealities of the forward-pass, but backward and update passes
    are ideal.

    During inference, statistical models of programming, drift
    and read noise can be used.
    """
    # pylint: disable=too-many-instance-attributes

    tile_class: ClassVar[Type] = InferenceTile
    """Tile class that correspond to this RPUConfig."""

    bindings_class: ClassVar[Type] = devices.AnalogTileParameter

    forward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the forward direction."""

    noise_model: BaseNoiseModel = field(default_factory=PCMLikeNoiseModel)
    """Statistical noise model to be used during (realistic) inference."""

    drift_compensation: BaseDriftCompensation = field(default_factory=GlobalDriftCompensation)
    """For compensating the drift during inference only."""

    clip: WeightClipParameter = field(default_factory=WeightClipParameter)
    """Parameters for weight clip."""

    modifier: WeightModifierParameter = field(default_factory=WeightModifierParameter)
    """Parameters for weight modifier."""

    # The following fields are not included in `__init__`, and should be
    # treated as read-only.

    device: IdealDevice = field(default_factory=IdealDevice,
                                init=False)
    """Parameters that modify the behavior of the pulsed device: ideal device."""

    backward: IOParameters = field(
        default_factory=lambda: IOParameters(is_perfect=True),
        init=False
    )
    """Input-output parameter setting for the backward direction: perfect."""

    update: UpdateParameters = field(
        default_factory=lambda: UpdateParameters(pulse_type=PulseType.NONE),
        init=False
    )
    """Parameter for the update behavior: ``NONE`` pulse type."""

    def as_bindings(self) -> devices.AnalogTileParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return tile_parameters_to_bindings(self)


@dataclass
class DigitalRankUpdateRPUConfig(_PrintableMixin):
    """Configuration for an analog (unit cell) resistive processing unit
    where the rank update is done in digital.

    Note that for forward and backward, an analog crossbar is still
    used, and during update the digitally computed rank update is
    transferred to the analog crossbar using pulses.
    """

    tile_class: ClassVar[Type] = AnalogTile
    """Tile class that correspond to this RPUConfig."""

    bindings_class: ClassVar[Type] = devices.AnalogTileParameter

    device: DigitalRankUpdateCell = field(default_factory=DigitalRankUpdateCell)
    """Parameters that modify the behavior of the pulsed device."""

    forward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the forward direction."""

    backward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the backward direction."""

    update: UpdateParameters = field(default_factory=UpdateParameters)
    """Parameter for the analog part of the update, that is the transfer
    from the digital buffer to the devices."""

    def as_bindings(self) -> devices.AnalogTileParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return tile_parameters_to_bindings(self)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Configurations presets for resistive processing units."""

from .configs import (
    # Single device configs.
    ReRamESPreset, ReRamSBPreset, CapacitorPreset, EcRamPreset, EcRamMOPreset, IdealizedPreset,
    GokmenVlasovPreset, PCMPreset,
    # 2-device configs.
    ReRamES2Preset, ReRamSB2Preset, Capacitor2Preset, EcRam2Preset, EcRamMO2Preset,
    Idealized2Preset,
    # 4-device configs.
    ReRamES4Preset, ReRamSB4Preset, Capacitor4Preset, EcRam4Preset, EcRamMO4Preset,
    Idealized4Preset,
    # Tiki-taka configs.
    TikiTakaReRamESPreset, TikiTakaReRamSBPreset, TikiTakaCapacitorPreset,
    TikiTakaEcRamPreset, TikiTakaEcRamMOPreset, TikiTakaIdealizedPreset,
    # MixedPrecision configs.
    MixedPrecisionReRamESPreset, MixedPrecisionReRamSBPreset, MixedPrecisionCapacitorPreset,
    MixedPrecisionEcRamPreset, MixedPrecisionEcRamMOPreset, MixedPrecisionIdealizedPreset,
    MixedPrecisionGokmenVlasovPreset, MixedPrecisionPCMPreset
)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""High level analog tiles (floating point)."""

from copy import deepcopy
from typing import Optional, Union, TYPE_CHECKING

from torch import device as torch_device
from torch.cuda import device as cuda_device

from aihwkit.exceptions import CudaError
from aihwkit.simulator.rpu_base import cuda, tiles
from aihwkit.simulator.tiles.base import BaseTile

if TYPE_CHECKING:
    from aihwkit.simulator.configs import FloatingPointRPUConfig


class FloatingPointTile(BaseTile):
    r"""Floating point tile.

    Implements a floating point or ideal analog tile.

    A linear layer with this tile is perfectly linear, it just uses
    the RPUCuda library for execution.

    **Forward pass**:

    .. math:: \mathbf{y} = W\mathbf{x}

    :math:`W` are the weights, :math:`\mathbf{x}` is the input
    vector. :math:`\mathbf{y}` is output of the vector matrix
    multiplication.  Note that if bias is used, :math:`\mathbf{x}` is
    concatenated with 1 so that the last column of :math:`W` are the
    biases.


    **Backward pass**:

    Typical backward pass with transposed weights:

    .. math:: \mathbf{d'} = W^T\mathbf{d}

    where :math:`\mathbf{d}` is the error
    vector. :math:`\mathbf{d}_o` is output of the backward matrix
    vector multiplication.


    **Weight update**:

    Usual learning rule for back-propagation:

    .. math:: w_{ij} \leftarrow w_{ij} + \lambda d_i\,x_j


    **Decay**:

    .. math:: w_{ij} \leftarrow w_{ij}(1-\alpha r_\text{decay})

    Weight decay can be called by calling the analog tile decay.

    Note:
        ``life_time`` parameter is set during
        initialization. alpha is a scaling factor that can be given
        during run-time.


    **Diffusion**:

    .. math::  w_{ij} \leftarrow w_{ij} +  \xi\;r_\text{diffusion}

    Similar to the decay, diffusion is only done when explicitly
    called. However, the parameter of the diffusion process are
    set during initialization and are fixed for the
    remainder. :math:`\xi` is a standard Gaussian process.

    Args:
        out_size: output vector size of the tile, ie. the dimension of
            :math:`\mathbf{y}` in case of :math:`\mathbf{y} =
            W\mathbf{x}` (or equivalently the dimension of the
            :math:`\boldsymbol{\delta}` of the backward pass).
        in_size: input vector size, ie. the dimension of the vector
            :math:`\mathbf{x}` in case of :math:`\mathbf{y} =
            W\mathbf{x}`).
        rpu_config: resistive processing unit configuration.
        bias: whether to add a bias column to the tile, ie. :math:`W`
            has an extra column to code the biases. Internally, the
            input :math:`\mathbf{x}` will be automatically expanded by
            an extra dimension which will be set to 1 always.
        in_trans: Whether to assume an transposed input (batch first).
        out_trans: Whether to assume an transposed output (batch first).
    """

    def __init__(
            self,
            out_size: int,
            in_size: int,
            rpu_config: Optional['FloatingPointRPUConfig'] = None,
            bias: bool = False,
            in_trans: bool = False,
            out_trans: bool = False,
    ):
        if not rpu_config:
            # Import `FloatingPointRPUConfig` dynamically to avoid import cycles.
            # pylint: disable=import-outside-toplevel
            from aihwkit.simulator.configs import FloatingPointRPUConfig
            rpu_config = FloatingPointRPUConfig()
        super().__init__(out_size, in_size, rpu_config, bias, in_trans, out_trans)

    def cpu(self) -> 'BaseTile':
        """Return a copy of this tile in CPU memory.

        Note:
            CUDA tiles weight can be accessed by `get_weights` etc
            methods, there is no need to move them to CPU and it is
            currently not supported.

        Returns:
            self in case of CPU

        Raises:
            CudaError: if a CUDA tile is moved to CPU
        """
        if self.is_cuda:
            raise CudaError('Currently it is not possible to move CUDA tile to cpu.')

        return self

    def cuda(
            self,
            device: Optional[Union[torch_device, str, int]] = None
    ) -> 'BaseTile':
        """Return a copy of this tile in CUDA memory.

        Args:
            device: CUDA device

        Returns:
            Self with the underlying C++ tile moved to CUDA memory.

        Raises:
            CudaError: if the library has not been compiled with CUDA.
        """
        if not cuda.is_compiled():
            raise CudaError('aihwkit has not been compiled with CUDA support')

        device = torch_device('cuda', cuda_device(device).idx)

        if self.is_cuda and device != self.device:
            raise CudaError('Cannot switch CUDA devices of existing Cuda tiles')

        if isinstance(self.tile, tiles.FloatingPointTile):
            with cuda_device(device):
                self.tile = tiles.CudaFloatingPointTile(self.tile)
                self.is_cuda = True
                self.device = device
                self.analog_ctx.cuda(device)

        return self

    def _create_simulator_tile(
            self,
            x_size: int,
            d_size: int,
            rpu_config: 'FloatingPointRPUConfig'
    ) -> tiles.FloatingPointTile:
        """Create a simulator tile.

        Args:
            x_size: input size
            d_size: output size
            rpu_config: resistive processing unit configuration

        Returns:
            a simulator tile based on the specified configuration.
        """
        meta_parameter = rpu_config.device.as_bindings()

        return meta_parameter.create_array(x_size, d_size)


class CudaFloatingPointTile(FloatingPointTile):
    """Floating point tile (CUDA).

    Floating point tile that uses GPU for its operation. The instantiation is
    based on an existing non-cuda tile: all the source attributes are copied
    except for the simulator tile, which is recreated using a GPU tile.

    Caution:
        Deprecated. Use ``FloatingPointTile(..).cuda()`` instead.

    Args:
        source_tile: tile to be used as the source of this tile
    """

    is_cuda = True

    def __init__(self, source_tile: FloatingPointTile):
        if not cuda.is_compiled():
            raise CudaError('aihwkit has not been compiled with CUDA support')

        # Create a new instance of the rpu config.
        new_rpu_config = deepcopy(source_tile.rpu_config)

        # Create the tile, replacing the simulator tile.
        super().__init__(source_tile.out_size, source_tile.in_size, new_rpu_config,
                         source_tile.bias, source_tile.in_trans, source_tile.out_trans)
        self.cuda(self.device)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Autograd functions for aihwkit."""

from typing import Any, Optional, Tuple

from torch import Tensor, empty_like
from torch.autograd import Function
from aihwkit.optim.context import AnalogContext


class AnalogFunctionBase(Function):
    """Base function for analog functions."""
    # pylint: disable=arguments-differ, protected-access

    @staticmethod
    def forward(
            ctx: Any,
            analog_ctx: AnalogContext,
            input_: Tensor,
            shared_weights: Optional[Tensor] = None,
            is_test: bool = False) -> Tensor:
        """Execute the forward pass in the analog tile.

        Note: Indexed versions can used when analog_ctx.use_indexed is
        set to True.
        """
        # Store in context for using during `backward()`.
        analog_tile = analog_ctx.analog_tile
        ctx.analog_ctx = analog_ctx
        ctx.shared_weights = None
        ctx.save_for_backward(input_)

        use_indexed = analog_ctx.use_indexed
        if shared_weights is not None:
            ctx.shared_weights = shared_weights
            analog_tile.ensure_shared_weights(shared_weights)
            analog_ctx.use_torch_update = True
        else:
            analog_ctx.use_torch_update = False

        # Invoke the forward pass in the tile instance.
        if use_indexed:
            return analog_tile.forward_indexed(input_, is_test)
        return analog_tile.forward(input_, is_test)

    @staticmethod
    def backward(
            ctx: Any,
            grad_output: Tensor,
    ) -> Tuple[Optional[Tensor], Optional[Tensor], Optional[Tensor], Optional[Tensor]]:
        """Execute the backward pass in the analog tile."""
        analog_ctx = ctx.analog_ctx
        analog_tile = analog_ctx.analog_tile
        input_, = ctx.saved_tensors

        shared_weights_grad = None
        use_indexed = analog_ctx.use_indexed

        if ctx.shared_weights is not None:
            analog_tile.ensure_shared_weights(ctx.shared_weights)

        # Call the backward function in the tile instance.
        if use_indexed:
            grad_input = analog_tile.backward_indexed(grad_output)
        else:
            grad_input = analog_tile.backward(grad_output)

        if analog_ctx.use_torch_update:
            # Grad computed directly (for inference training)
            shared_weights_grad = empty_like(ctx.shared_weights)
            analog_tile.set_delta_weights(shared_weights_grad)
            if use_indexed:
                analog_tile.update_indexed(input_, grad_output)
            else:
                analog_tile.update(input_, grad_output)
            analog_tile.reset_delta_weights()
        else:
            # Store activation and errors for optimizer (for analog training)
            analog_ctx.analog_input.append(input_)
            analog_ctx.analog_grad_output.append(grad_output)

        return None, grad_input, shared_weights_grad, None


class AnalogFunction(AnalogFunctionBase):
    """Function that delegates into a `RPU` unit."""
    # pylint: disable=arguments-differ

    @staticmethod
    def forward(
            ctx: Any,
            analog_ctx: AnalogContext,
            input_: Tensor,
            shared_weights: Optional[Tensor] = None,
            is_test: bool = False) -> Tensor:
        """Execute the forward pass in the analog tile."""
        analog_ctx.use_indexed = False
        return AnalogFunctionBase.forward(
            ctx, analog_ctx, input_, shared_weights, is_test)


class AnalogIndexedFunction(AnalogFunctionBase):
    """Function that delegates into a `RPU` unit to use the indexed forward/backward/update."""
    # pylint: disable=arguments-differ

    @staticmethod
    def forward(
            ctx: Any,
            analog_ctx: AnalogContext,
            input_: Tensor,
            shared_weights: Optional[Tensor] = None,
            is_test: bool = False) -> Tensor:
        """Execute the forward pass in the analog tile."""
        analog_ctx.use_indexed = True
        return AnalogFunctionBase.forward(
            ctx, analog_ctx, input_, shared_weights, is_test)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Digital/analog model conversion utilities.

This module includes tools for converting a given torch model to a
model containing analog layers.
"""

from typing import TypeVar, Optional, Dict
import copy

from torch.nn import Module, Linear, Conv1d, Conv2d, Conv3d, Sequential

from aihwkit.nn import (
    AnalogLinear, AnalogConv1d, AnalogConv2d, AnalogConv3d, AnalogSequential
)

RPUConfigGeneric = TypeVar('RPUConfigGeneric')

_DEFAULT_CONVERSION_MAP = {Linear: AnalogLinear,
                           Conv1d: AnalogConv1d,
                           Conv2d: AnalogConv2d,
                           Conv3d: AnalogConv3d,
                           Sequential: AnalogSequential}


def convert_to_analog(
        module: Module,
        rpu_config: RPUConfigGeneric,
        realistic_read_write: bool = False,
        weight_scaling_omega: float = 0.0,
        conversion_map: Optional[Dict] = None
) -> Module:
    """Convert a given digital model to analog counter parts.

    Note:
        The torch device (cuda/cpu) is inferred from the original
        models parameters, however, if multiple torch
        devices are used in a given module, the corresponding analog
        module is not moved to any device.

    Args:
        module: The torch module to convert. All layers that are
            defined in the ``conversion_map``.
        rpu_config: RPU config to apply to all converted tiles.
            Applied to all converted tiles.
        realistic_read_write: Whether to use closed-loop programming
            when setting the weights. Applied to all converted tiles.
        weight_scaling_omega: If non-zero, applied weights of analog
            layers will be scaled by ``weight_scaling_omega`` divided by
            the absolute maximum value of the original weight matrix.

            Note:
                Make sure that the weight max and min setting of the
                device support the desired analog weight range.
        conversion_map: Dictionary of module classes to be replaced in
            case of custom replacement rules. By default all ``Conv`` and ``Linear``
            layers are replaced with their analog counterparts.

            Note:
                The analog layer needs to have a class method
                ``from_digital`` which will be called during the
                conversion.

    Returns:
        module with replaced digital layers with analog layers.
    """
    module = copy.deepcopy(module)

    if conversion_map is None:
        conversion_map = _DEFAULT_CONVERSION_MAP

    # Convert parent.
    if module.__class__ in conversion_map:
        module = conversion_map[module.__class__].from_digital(  # type: ignore
            module, rpu_config, realistic_read_write, weight_scaling_omega)

    # Convert children.
    convert_dic = {}
    for name, mod in module.named_children():

        n_grand_children = len(list(mod.named_children()))
        if n_grand_children > 0:
            new_mod = convert_to_analog(mod, rpu_config, realistic_read_write,
                                        weight_scaling_omega, conversion_map)

        elif mod.__class__ in conversion_map:
            new_mod = conversion_map[mod.__class__].from_digital(   # type: ignore
                mod, rpu_config, realistic_read_write, weight_scaling_omega)
        else:
            continue

        devices = {p.device for p in mod.parameters()}
        if len(devices) == 1:
            # We only use "to" if device is unique.
            new_mod = new_mod.to(list(devices)[0])
        convert_dic[name] = new_mod

    for name, new_mod in convert_dic.items():
        module._modules[name] = new_mod  # pylint: disable=protected-access

    return module



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Parsers for the AIHW Composer API."""

from datetime import datetime, timezone
from typing import Any, Dict

from aihwkit.cloud.client.entities import (
    CloudExperiment, CloudExperimentCategory, CloudJobStatus, CloudJob
)
from aihwkit.cloud.client.exceptions import InvalidResponseFieldError


class ExperimentParser:
    """Parser for Experiment API responses."""

    @staticmethod
    def parse_experiment(api_response: Dict, api_client: Any) -> CloudExperiment:
        """Return a CloudExperiment from an API response.

        Args:
            api_response: the response from the API.
            api_client: the client to be used in API requests.

        Returns:
            A `CloudExperiment` based on the response. Some of the fields might
            not be populated if they are not present in the response.
        """
        experiment = CloudExperiment(
            _api_client=api_client,
            id_=api_response['id'],
            name=api_response['name'],
            category=ExperimentParser.parse_experiment_category(api_response),
            created_at=ExperimentParser.parse_date_string(api_response['createdAt']),
            input_id=None,
            job=None
        )

        if api_response.get('input', {}).get('id'):
            experiment.input_id = api_response['input']['id']

        if api_response.get('job', None):
            experiment.job = ExperimentParser.parse_job(api_response['job'])

        return experiment

    @staticmethod
    def parse_job(api_response: Dict) -> CloudJob:
        """Return an CloudJob from an API response.

        Args:
            api_response: the response from the API.

        Returns:
            A `CloudJob` based on the response. Some of the fields might
            not be populated if they are not present in the response.
        """
        job = CloudJob(
            id_=api_response['id'],
            output_id=None,
            status=ExperimentParser.parse_experiment_status(api_response)
        )

        if api_response.get('output', None):
            job.output_id = api_response['output']

        return job

    @staticmethod
    def parse_experiment_status(api_response: Dict) -> CloudJobStatus:
        """Return an Experiment status from an API response.

        Args:
            api_response: the response from the API.

        Returns:
            A value from the `CloudJobStatus` enum.

        Raises:
            InvalidResponseFieldError: if the API response contains an
                unrecognized status code.
        """
        job_status = api_response['status']

        if job_status in ('waiting', 'validating', 'validated'):
            return CloudJobStatus.WAITING
        if job_status in ('running', ):
            return CloudJobStatus.RUNNING
        if job_status in ('failed', 'cancelled'):
            return CloudJobStatus.FAILED
        if job_status in ('completed',):
            return CloudJobStatus.COMPLETED

        raise InvalidResponseFieldError('Unsupported job status: {}'.format(job_status))

    @staticmethod
    def parse_experiment_category(api_response: Dict) -> CloudExperimentCategory:
        """Return an Experiment category from an API response.

        Args:
            api_response: the response from the API.

        Returns:
            A value from the `CloudExperimentCategory` enum.

        Raises:
            InvalidResponseFieldError: if the API response contains an
                unrecognized category.
        """
        job_category = api_response['category']

        if job_category in ('train', 'trainweb'):
            return CloudExperimentCategory.BASIC_TRAINING

        raise InvalidResponseFieldError(
            'Unsupported experiment category: {}'.format(job_category))

    @staticmethod
    def parse_date_string(date_string: str) -> datetime:
        """Return a datetime from a date string.

        Args:
            date_string: the date string from the API.

        Returns:
            A value from the `CloudExperimentCategory` enum.
        """
        tmp_datetime = datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S.%fZ')

        return tmp_datetime.replace(tzinfo=timezone.utc)


class GeneralParser:
    """Parser for generic responses."""
    # pylint: disable=too-few-public-methods

    @staticmethod
    def parse_login(api_response: Dict) -> str:
        """Return the jwt token from an API response.

        Args:
            api_response: the response from the API.

        Returns:
            A string with the jwt token.
        """
        return api_response['jwt']



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# This file has been adapted from:
# https://github.com/pytorch/pytorch/blob/master/benchmarks/fastrnns/custom_lstms.py
# Licensed under the following terms (https://github.com/pytorch/pytorch/blob/master/LICENSE):

# From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou,
#                                                   Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the names of Facebook, Deepmind Technologies, NYU, NEC Laboratories America
#    and IDIAP Research Institute nor the names of its contributors may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Analog LSTM layers."""

import warnings
import math
from typing import Any, List, Optional, Tuple, Type, Callable
from collections import namedtuple

from torch import Tensor, sigmoid, stack, tanh, jit, zeros
from torch.nn import Dropout, ModuleList, init

from aihwkit.nn import AnalogLinear, AnalogSequential
from aihwkit.nn.modules.base import RPUConfigAlias
from aihwkit.simulator.configs import SingleRPUConfig
from aihwkit.simulator.configs.devices import ConstantStepDevice
from aihwkit.nn.modules.base import AnalogModuleBase


LSTMState = namedtuple('LSTMState', ['hx', 'cx'])


class AnalogLSTMCell(AnalogSequential):
    """Analog LSTM Cell.

    Args:
        input_size: in_features size for W_ih matrix
        hidden_size: in_features and out_features size for W_hh matrix
        bias: whether to use a bias row on the analog tile or not
        rpu_config: configuration for an analog resistive processing unit
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights
        weight_scaling_omega: the weight value where the max
            weight will be scaled to. If zero, no weight scaling will
            be performed
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            bias: bool,
            rpu_config: Optional[RPUConfigAlias],
            realistic_read_write: bool = False,
            weight_scaling_omega: float = 0.0
    ):
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.weight_ih = AnalogLinear(input_size, 4 * hidden_size, bias=bias,
                                      rpu_config=rpu_config,
                                      realistic_read_write=realistic_read_write,
                                      weight_scaling_omega=weight_scaling_omega)
        self.weight_hh = AnalogLinear(hidden_size, 4 * hidden_size, bias=bias,
                                      rpu_config=rpu_config,
                                      realistic_read_write=realistic_read_write,
                                      weight_scaling_omega=weight_scaling_omega)

    def forward(
            self,
            input_: Tensor,
            state: Tuple[Tensor, Tensor]
    ) -> Tuple[Tensor, Tuple[Tensor, Tensor]]:
        # pylint: disable=arguments-differ
        h_x, c_x = state
        gates = self.weight_ih(input_) + self.weight_hh(h_x)
        in_gate, forget_gate, cell_gate, out_gate = gates.chunk(4, 1)

        in_gate = sigmoid(in_gate)
        forget_gate = sigmoid(forget_gate)
        cell_gate = tanh(cell_gate)
        out_gate = sigmoid(out_gate)

        c_y = (forget_gate * c_x) + (in_gate * cell_gate)
        h_y = out_gate * tanh(c_y)
        return h_y, (h_y, c_y)


class AnalogLSTMLayer(AnalogSequential):
    """Analog LSTM Layer.

    Args:
        cell: LSTMCell type (e.g. AnalogLSTMCell)
        cell_args: arguments to LSTMCell (e.g. input_size, hidden_size, rpu_configs)
    """
    # pylint: disable=abstract-method

    def __init__(self, cell: AnalogLSTMCell, *cell_args: Any):
        super().__init__()
        self.cell = cell(*cell_args)

    def forward(
            self, input_: Tensor,
            state: Tuple[Tensor, Tensor]
    ) -> Tuple[Tensor, Tuple[Tensor, Tensor]]:
        # pylint: disable=arguments-differ
        inputs = input_.unbind(0)
        outputs = jit.annotate(List[Tensor], [])
        for input_item in inputs:
            out, state = self.cell(input_item, state)
            outputs += [out]
        return stack(outputs), state


class ModularAnalogLSTMWithDropout(AnalogSequential):
    """Modular Analog LSTM with dropout.

    Args:
        num_layers: number of serially connected LSTM layers
        layer: LSTM layer type (e.g. AnalogLSTMLayer)
        dropout: dropout applied to output of all LSTM layers except last
        first_layer_args: LSTMCell type, input_size, hidden_size, rpu_config, etc.
        other_layer_args: LSTMCell type, hidden_size, hidden_size, rpu_config, etc.
    """
    # pylint: disable=abstract-method

    # Necessary for iterating through self.layers and dropout support
    __constants__ = ['layers', 'num_layers']

    def __init__(
            self,
            num_layers: int,
            layer: Type,
            dropout: float,
            first_layer_args: Any,
            other_layer_args: Any):
        super().__init__()
        self.layers = init_stacked_analog_lstm(num_layers, layer, first_layer_args,
                                               other_layer_args)

        # Introduce a Dropout layer on the outputs of each LSTM layer except
        # the last layer.
        self.num_layers = num_layers
        if num_layers == 1 and dropout > 0:
            warnings.warn('dropout lstm adds dropout layers after all but last '
                          'recurrent layer, it expects num_layers greater than '
                          '1, but got num_layers = 1')
        self.dropout_layer = Dropout(dropout) if dropout else None

    def forward(
            self,
            input_: Tensor,
            states: List[Tuple[Tensor, Tensor]]
    ) -> Tuple[Tensor, List[Tuple[Tensor, Tensor]]]:
        # pylint: disable=arguments-differ
        # List[LSTMState]: One state per layer.
        output_states = jit.annotate(List[Tuple[Tensor, Tensor]], [])
        output = input_

        for i, rnn_layer in enumerate(self.layers):
            state = states[i]
            output, out_state = rnn_layer(output, state)
            # Apply the dropout layer except the last layer.
            if i < self.num_layers - 1 and self.dropout_layer is not None:
                output = self.dropout_layer(output)
            output_states += [out_state]
            i += 1

        return output, output_states


class AnalogLSTM(AnalogSequential):
    """Modular LSTM that uses analog tiles.

    Args:
        input_size: in_features to W_{ih} matrix of first layer
        hidden_size: in_features and out_features for W_{hh} matrices
        num_layers: number of serially connected LSTM layers
        dropout: dropout applied to output of all LSTM layers except last
        bias: whether to use a bias row on the analog tile or not
        rpu_config: resistive processing unit configuration.
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights
        weight_scaling_omega: the weight value where the max
            weight will be scaled to. If zero, no weight scaling will
            be performed
        xavier: whether standard PyTorch LSTM weight
            initialization (default) or Xavier initialization
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            input_size: int,
            hidden_size: int,
            num_layers: int = 1,
            dropout: float = 0.0,
            bias: bool = True,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: float = 0.0,
            xavier: bool = False):
        super().__init__()

        # Default to SingleRPUConfig with ConstantStepDevice.
        if not rpu_config:
            rpu_config = SingleRPUConfig(device=ConstantStepDevice())

        self.lstm = ModularAnalogLSTMWithDropout(
            num_layers, AnalogLSTMLayer, dropout,
            first_layer_args=[AnalogLSTMCell, input_size, hidden_size, bias,
                              rpu_config, realistic_read_write, weight_scaling_omega],
            other_layer_args=[AnalogLSTMCell, hidden_size, hidden_size, bias,
                              rpu_config, realistic_read_write, weight_scaling_omega])
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.reset_parameters(xavier)

    def init_layers(
            self,
            weight_init_fn: Callable,
            bias_init_fn: Optional[Callable] = None
    ) -> None:
        """Init the analog layers with custom functions.

        Args:
            weight_init_fn: in-place tensor function applied to weight of
                ``AnalogLinear`` layers
            bias_init_fn: in-place tensor function applied to bias of
                ``AnalogLinear`` layers

        Note:
            If no bias init function is provided the weight init
            function is taken for the bias as well.
        """
        def init_analog_layer(layer: AnalogModuleBase) -> None:
            """Init the weights and bias of an analog linear layer."""
            weight_init_fn(layer.weight.data)
            if layer.use_bias:
                if bias_init_fn is None:
                    weight_init_fn(layer.bias.data)
                else:
                    bias_init_fn(layer.bias.data)

            layer.set_weights(layer.weight, layer.bias)

        self._apply_to_analog(init_analog_layer)  # pylint: disable=protected-access

    def reset_parameters(self, xavier: bool = False) -> None:
        """Weight and bias initialization.

        Args:
            xavier: whether standard PyTorch LSTM weight
               initialization (default) or Xavier initialization
        """
        if xavier:
            self.init_layers(init.xavier_uniform_, init.zeros_)
        else:
            stdv = 1. / math.sqrt(self.hidden_size)
            self.init_layers(lambda x: x.uniform_(-stdv, stdv))

    def forward(
            self,
            x: Tensor,
            states: Optional[List[Tuple[Tensor, Tensor]]] = None
    ) -> Tuple[Tensor, List[Tuple[Tensor, Tensor]]]:
        # pylint: disable=arguments-differ
        if states is None:
            # TODO: batch_first.
            states = [LSTMState(zeros(x.size()[1], self.hidden_size),
                                zeros(x.size()[1], self.hidden_size))
                      for _ in range(self.num_layers)]

        return self.lstm(x, states)


def init_stacked_analog_lstm(
        num_layers: int,
        layer: Type,
        first_layer_args: Any,
        other_layer_args: Any
) -> ModuleList:
    """Construct a list of LSTMLayers over which to iterate.

    Args:
        num_layers: number of serially connected LSTM layers
        layer: LSTM layer type (e.g. AnalogLSTMLayer)
        first_layer_args: LSTMCell type, input_size, hidden_size, rpu_config, etc.
        other_layer_args: LSTMCell type, hidden_size, hidden_size, rpu_config, etc.

    Returns:
        torch.nn.ModuleList, which is similar to a regular Python list,
        but where torch.nn.Module methods can be applied
    """
    layers = [layer(*first_layer_args)] \
        + [layer(*other_layer_args) for _ in range(num_layers - 1)]
    return ModuleList(layers)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Utilities for resistive processing units configurations."""

from dataclasses import Field, fields, is_dataclass
from enum import Enum
from textwrap import indent
from typing import Any, List

from aihwkit.simulator.rpu_base import devices, tiles


def parameters_to_bindings(params: Any) -> Any:
    """Convert a dataclass parameter into a bindings class."""
    result = params.bindings_class()
    for field, value in params.__dict__.items():
        # Convert enums to the bindings enums.
        if field in ('unit_cell_devices', 'device'):
            # Exclude special fields that are not present in the bindings.
            continue

        if isinstance(value, Enum):
            if hasattr(tiles, value.__class__.__name__):
                enum_class = getattr(tiles, value.__class__.__name__)
            else:
                enum_class = getattr(devices, value.__class__.__name__)
            enum_value = getattr(enum_class, value.value)
            setattr(result, field, enum_value)
        elif is_dataclass(value):
            setattr(result, field, parameters_to_bindings(value))
        else:
            setattr(result, field, value)

    return result


def tile_parameters_to_bindings(params: Any) -> Any:
    """Convert a tile dataclass parameter into a bindings class."""
    field_map = {'forward': 'forward_io',
                 'backward': 'backward_io'}
    excluded_fields = ('device', 'noise_model', 'drift_compensation',
                       'clip', 'modifier')

    result = params.bindings_class()
    for field, value in params.__dict__.items():
        # Get the mapped field name, if needed.
        field = field_map.get(field, field)

        # Convert enums to the bindings enums.
        if field in excluded_fields:
            # Exclude special fields that are not present in the bindings.
            continue

        if isinstance(value, Enum):
            enum_class = getattr(devices, value.__class__.__name__)
            enum_value = getattr(enum_class, value.value)
            setattr(result, field, enum_value)
        elif is_dataclass(value):
            setattr(result, field, parameters_to_bindings(value))
        else:
            setattr(result, field, value)

    return result


class _PrintableMixin:
    """Helper class for pretty-printing of config dataclasses."""
    # pylint: disable=too-few-public-methods

    def __str__(self) -> str:
        """Return a pretty-print representation."""

        def lines_list_to_str(
                lines_list: List[str],
                prefix: str = '',
                suffix: str = '',
                indent_: int = 0,
                force_multiline: bool = False
        ) -> str:
            """Convert a list of lines into a string.

            Args:
                lines_list: the list of lines to be converted.
                prefix: an optional prefix to be appended at the beginning of
                    the string.
                suffix: an optional suffix to be appended at the end of the string.
                indent_: the optional number of spaces to indent the code.
                force_multiline: force the output to be multiline.

            Returns:
                The lines collapsed into a single string (potentially with line
                breaks).
            """
            if force_multiline or len(lines_list) > 3 or any(
                    '\n' in line for line in lines_list):
                # Return a multi-line string.
                lines_str = indent(',\n'.join(lines_list), ' '*indent_)
                prefix = '{}\n'.format(prefix) if prefix else prefix
                suffix = '\n{}'.format(suffix) if suffix else suffix
            else:
                # Return an inline string.
                lines_str = ', '.join(lines_list)

            return '{}{}{}'.format(prefix, lines_str, suffix)

        def field_to_str(field_value: Any) -> str:
            """Return a string representation of the value of a field.

            Args:
                field_value: the object that contains a field value.

            Returns:
                The string representation of the field (potentially with line
                breaks).
            """
            field_lines = []
            force_multiline = False

            # Handle special cases.
            if isinstance(field_value, list) and len(value) > 0:
                # For non-emtpy lists, always use multiline, with one item per line.
                for item in field_value:
                    field_lines.append(indent('{}'.format(str(item)), ' '*4))
                force_multiline = True
            else:
                field_lines.append(str(field_value))

            prefix = '[' if force_multiline else ''
            suffix = ']' if force_multiline else ''
            return lines_list_to_str(
                field_lines, prefix, suffix, force_multiline=force_multiline)

        def is_skippable(field: Field, value: Any) -> bool:
            """Return whether a field should be skipped."""
            if value == field.default:
                # Skip fields with the default value.
                return True

            if 'hide_if' in field.metadata and field.metadata.get('hide_if') == value:
                return True

            return False

        # Main loop.

        # Build the list of lines.
        fields_lines = []
        for field in fields(self):
            value = getattr(self, field.name)

            # Exclude fields.
            if is_skippable(field, value):
                continue

            # Convert the value into a string, falling back to repr if needed.
            try:
                value_str = field_to_str(value)
            except Exception:  # pylint: disable=broad-except
                value_str = str(value)

            fields_lines.append('{}={}'.format(field.name, value_str))

        # Convert the full object to str.
        output = lines_list_to_str(
            fields_lines, '{}('.format(self.__class__.__name__), ')', 4)

        return output



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for layer abstractions."""

from torch import randn
from torch.nn import (Conv1d as torch_Conv1d, Conv2d as torch_Conv2d,
                      Conv3d as torch_Conv3d, Sequential)
from torch.nn.functional import mse_loss

from aihwkit.optim import AnalogSGD

from .helpers.decorators import parametrize_over_layers
from .helpers.layers import Conv1d, Conv1dCuda, Conv2d, Conv2dCuda, Conv3d, Conv3dCuda
from .helpers.testcases import ParametrizedTestCase
from .helpers.tiles import FloatingPoint, Inference


class ConvolutionLayerTest(ParametrizedTestCase):
    """Generic class for helping testing analog convolution layers."""

    digital_layer_cls = torch_Conv1d

    def get_digital_layer(self, in_channels=2, out_channels=3, kernel_size=4, padding=2):
        """Return a digital layer."""
        layer = self.digital_layer_cls(in_channels=in_channels,
                                       out_channels=out_channels,
                                       kernel_size=kernel_size,
                                       padding=padding,
                                       bias=self.bias)
        if self.use_cuda:
            layer = layer.cuda()

        return layer

    def set_weights_from_digital_model(self, analog_model, digital_model):
        """Set the analog model weights based on the digital model."""
        weights, biases = self.get_weights_from_digital_model(analog_model, digital_model)
        analog_model.analog_tile.set_weights(weights, biases, realistic=False)

    @staticmethod
    def get_weights_from_digital_model(analog_model, digital_model):
        """Set the analog model weights based on the digital model."""
        weights = digital_model.weight.data.detach().reshape(
            [analog_model.out_features, analog_model.in_features]).cpu()
        biases = None
        if digital_model.bias is not None:
            biases = digital_model.bias.data.detach().cpu()

        return weights, biases

    @staticmethod
    def train_model(model, loss_func, x_b, y_b):
        """Train the model."""
        opt = AnalogSGD(model.parameters(), lr=0.1)
        opt.regroup_param_groups(model)

        epochs = 10
        for _ in range(epochs):
            opt.zero_grad()
            pred = model(x_b)
            loss = loss_func(pred, y_b)
            loss.backward()
            opt.step()


@parametrize_over_layers(
    layers=[Conv1d, Conv1dCuda],
    tiles=[FloatingPoint, Inference],
    biases=[True, False]
)
class Convolution1dLayerTest(ConvolutionLayerTest):
    """Tests for AnalogConv1d layer."""

    digital_layer_cls = torch_Conv1d

    def test_torch_original_layer(self):
        """Test a single layer, having the digital layer as reference."""
        # This tests the forward pass
        model = self.get_digital_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        x = randn(3, 2, 4)

        if self.use_cuda:
            x = x.cuda()

        y = model(x)

        analog_model = self.get_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        self.set_weights_from_digital_model(analog_model, model)

        y_analog = analog_model(x)
        self.assertTensorAlmostEqual(y_analog, y)

    def test_torch_train_original_layer(self):
        """Test the forward and update pass, having the digital layer as reference."""
        model = self.get_digital_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        analog_model = self.get_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        self.set_weights_from_digital_model(analog_model, model)

        loss_func = mse_loss
        y_b = randn(3, 3, 5)
        x_b = randn(3, 2, 4)

        if self.use_cuda:
            y_b = y_b.cuda()
            x_b = x_b.cuda()

        self.train_model(model, loss_func, x_b, y_b)
        self.train_model(analog_model, loss_func, x_b, y_b)

        weight, bias = self.get_weights_from_digital_model(analog_model, model)

        weight_analog, bias_analog = analog_model.analog_tile.get_weights(realistic=False)

        self.assertTensorAlmostEqual(weight_analog, weight)
        if analog_model.use_bias:
            self.assertTensorAlmostEqual(bias_analog, bias)

    def test_torch_train_original_layer_multiple(self):
        """Test the backward pass, having the digital layer as reference."""
        model = Sequential(
            self.get_digital_layer(in_channels=2, out_channels=2, kernel_size=4, padding=2),
            self.get_digital_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        )

        analog_model = Sequential(
            self.get_layer(in_channels=2, out_channels=2, kernel_size=4, padding=2),
            self.get_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        )

        for analog_layer, layer in zip(analog_model.children(), model.children()):
            self.set_weights_from_digital_model(analog_layer, layer)

        loss_func = mse_loss
        y_b = randn(3, 3, 6)
        x_b = randn(3, 2, 4)

        if self.use_cuda:
            y_b = y_b.cuda()
            x_b = x_b.cuda()

        self.train_model(model, loss_func, x_b, y_b)
        self.train_model(analog_model, loss_func, x_b, y_b)

        for analog_layer, layer in zip(analog_model.children(), model.children()):
            weight, bias = self.get_weights_from_digital_model(analog_layer, layer)

            weight_analog, bias_analog = analog_layer.analog_tile.get_weights(realistic=False)

            self.assertTensorAlmostEqual(weight_analog, weight)
            if analog_layer.use_bias:
                self.assertTensorAlmostEqual(bias_analog, bias)

    def test_layer_instantiation(self):
        """Test AnalogConv2d layer instantiation."""
        model = self.get_layer(in_channels=2, out_channels=3, kernel_size=4)

        # Assert the number of elements of the weights.
        tile_weights, tile_biases = model.analog_tile.get_weights()

        self.assertEqual(tile_weights.numel(), 2*3*4)
        if model.use_bias:
            self.assertEqual(tile_biases.numel(), 3)


@parametrize_over_layers(
    layers=[Conv2d, Conv2dCuda],
    tiles=[FloatingPoint, Inference],
    biases=[True, False]
)
class Convolution2dLayerTest(ConvolutionLayerTest):
    """Tests for AnalogConv2d layer."""

    digital_layer_cls = torch_Conv2d

    def test_torch_original_layer(self):
        """Test a single layer, having the digital layer as reference."""
        # This tests the forward pass
        model = self.get_digital_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        x = randn(3, 2, 4, 4)

        if self.use_cuda:
            x = x.cuda()

        y = model(x)

        analog_model = self.get_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        self.set_weights_from_digital_model(analog_model, model)

        y_analog = analog_model(x)
        self.assertTensorAlmostEqual(y_analog, y)

    def test_torch_train_original_layer(self):
        """Test the forward and update pass, having the digital layer as reference."""
        model = self.get_digital_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        analog_model = self.get_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        self.set_weights_from_digital_model(analog_model, model)

        loss_func = mse_loss
        y_b = randn(3, 3, 5, 5)
        x_b = randn(3, 2, 4, 4)

        if self.use_cuda:
            y_b = y_b.cuda()
            x_b = x_b.cuda()

        self.train_model(model, loss_func, x_b, y_b)
        self.train_model(analog_model, loss_func, x_b, y_b)

        weight, bias = self.get_weights_from_digital_model(analog_model, model)

        weight_analog, bias_analog = analog_model.analog_tile.get_weights(realistic=False)

        self.assertTensorAlmostEqual(weight_analog, weight)
        if analog_model.use_bias:
            self.assertTensorAlmostEqual(bias_analog, bias)

    def test_torch_train_original_layer_multiple(self):
        """Test the backward pass, having the digital layer as reference."""
        model = Sequential(
            self.get_digital_layer(in_channels=2, out_channels=2, kernel_size=4, padding=2),
            self.get_digital_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        )

        analog_model = Sequential(
            self.get_layer(in_channels=2, out_channels=2, kernel_size=4, padding=2),
            self.get_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        )

        for analog_layer, layer in zip(analog_model.children(), model.children()):
            self.set_weights_from_digital_model(analog_layer, layer)

        loss_func = mse_loss
        y_b = randn(3, 3, 6, 6)
        x_b = randn(3, 2, 4, 4)

        if self.use_cuda:
            y_b = y_b.cuda()
            x_b = x_b.cuda()

        self.train_model(model, loss_func, x_b, y_b)
        self.train_model(analog_model, loss_func, x_b, y_b)

        for analog_layer, layer in zip(analog_model.children(), model.children()):
            weight, bias = self.get_weights_from_digital_model(analog_layer, layer)

            weight_analog, bias_analog = analog_layer.analog_tile.get_weights(realistic=False)

            self.assertTensorAlmostEqual(weight_analog, weight)
            if analog_layer.use_bias:
                self.assertTensorAlmostEqual(bias_analog, bias)

    def test_layer_instantiation(self):
        """Test AnalogConv2d layer instantiation."""
        model = self.get_layer(in_channels=2, out_channels=3, kernel_size=4)

        # Assert the number of elements of the weights.
        tile_weights, tile_biases = model.analog_tile.get_weights()

        self.assertEqual(tile_weights.numel(), 2*3*4*4)
        if model.use_bias:
            self.assertEqual(tile_biases.numel(), 3)


@parametrize_over_layers(
    layers=[Conv3d, Conv3dCuda],
    tiles=[FloatingPoint, Inference],
    biases=[True, False]
)
class Convolution3dLayerTest(ConvolutionLayerTest):
    """Tests for AnalogConv3d layer."""

    digital_layer_cls = torch_Conv3d

    def test_torch_original_layer(self):
        """Test a single layer, having the digital layer as reference."""
        # This tests the forward pass
        model = self.get_digital_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        x = randn(3, 2, 4, 5, 6)

        if self.use_cuda:
            x = x.cuda()

        y = model(x)

        analog_model = self.get_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        self.set_weights_from_digital_model(analog_model, model)

        y_analog = analog_model(x)
        self.assertTensorAlmostEqual(y_analog, y)

    def test_torch_train_original_layer(self):
        """Test the forward and update pass, having the digital layer as reference."""
        model = self.get_digital_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        analog_model = self.get_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        self.set_weights_from_digital_model(analog_model, model)

        loss_func = mse_loss
        y_b = randn(3, 3, 5, 5, 5)
        x_b = randn(3, 2, 4, 4, 4)

        if self.use_cuda:
            y_b = y_b.cuda()
            x_b = x_b.cuda()

        self.train_model(model, loss_func, x_b, y_b)
        self.train_model(analog_model, loss_func, x_b, y_b)

        weight, bias = self.get_weights_from_digital_model(analog_model, model)

        weight_analog, bias_analog = analog_model.analog_tile.get_weights(realistic=False)

        self.assertTensorAlmostEqual(weight_analog, weight)
        if analog_model.use_bias:
            self.assertTensorAlmostEqual(bias_analog, bias)

    def test_torch_train_original_layer_multiple(self):
        """Test the backward pass, having the digital layer as reference."""
        model = Sequential(
            self.get_digital_layer(in_channels=2, out_channels=2, kernel_size=4, padding=2),
            self.get_digital_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        )

        analog_model = Sequential(
            self.get_layer(in_channels=2, out_channels=2, kernel_size=4, padding=2),
            self.get_layer(in_channels=2, out_channels=3, kernel_size=4, padding=2)
        )

        for analog_layer, layer in zip(analog_model.children(), model.children()):
            self.set_weights_from_digital_model(analog_layer, layer)

        loss_func = mse_loss
        y_b = randn(3, 3, 6, 6, 6)
        x_b = randn(3, 2, 4, 4, 4)

        if self.use_cuda:
            y_b = y_b.cuda()
            x_b = x_b.cuda()

        self.train_model(model, loss_func, x_b, y_b)
        self.train_model(analog_model, loss_func, x_b, y_b)

        for analog_layer, layer in zip(analog_model.children(), model.children()):
            weight, bias = self.get_weights_from_digital_model(analog_layer, layer)

            weight_analog, bias_analog = analog_layer.analog_tile.get_weights(realistic=False)

            self.assertTensorAlmostEqual(weight_analog, weight)
            if analog_layer.use_bias:
                self.assertTensorAlmostEqual(bias_analog, bias)

    def test_layer_instantiation(self):
        """Test AnalogConv2d layer instantiation."""
        model = self.get_layer(in_channels=2, out_channels=3, kernel_size=4)

        # Assert the number of elements of the weights.
        tile_weights, tile_biases = model.analog_tile.get_weights()

        self.assertEqual(tile_weights.numel(), 2*3*4*4*4)
        if model.use_bias:
            self.assertEqual(tile_biases.numel(), 3)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import onnx_common_pb2 as onnx__common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='common.proto',
  package='aihwx',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0c\x63ommon.proto\x12\x05\x61ihwx\x1a\x11onnx_common.proto\"F\n\x0ePrimitiveProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"F\n\x0eTransformProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"O\n\x17\x41\x63tivationFunctionProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"m\n\nLayerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\x12)\n\nstate_dict\x18\x03 \x03(\x0b\x32\x15.aihwx.AttributeProto\"\x86\x01\n\x19LayerOrActivationFunction\x12=\n\x13\x61\x63tivation_function\x18\x01 \x01(\x0b\x32\x1e.aihwx.ActivationFunctionProtoH\x00\x12\"\n\x05layer\x18\x02 \x01(\x0b\x32\x11.aihwx.LayerProtoH\x00\x42\x06\n\x04item\"F\n\x0eOptimizerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"F\n\x0eSchedulerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"I\n\x11LossFunctionProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\";\n\x07Network\x12\x30\n\x06layers\x18\x01 \x03(\x0b\x32 .aihwx.LayerOrActivationFunction\"(\n\x07Version\x12\x0e\n\x06schema\x18\x01 \x02(\x03\x12\r\n\x05opset\x18\x02 \x02(\x03'
  ,
  dependencies=[onnx__common__pb2.DESCRIPTOR,])




_PRIMITIVEPROTO = _descriptor.Descriptor(
  name='PrimitiveProto',
  full_name='aihwx.PrimitiveProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.PrimitiveProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.PrimitiveProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42,
  serialized_end=112,
)


_TRANSFORMPROTO = _descriptor.Descriptor(
  name='TransformProto',
  full_name='aihwx.TransformProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.TransformProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.TransformProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=114,
  serialized_end=184,
)


_ACTIVATIONFUNCTIONPROTO = _descriptor.Descriptor(
  name='ActivationFunctionProto',
  full_name='aihwx.ActivationFunctionProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.ActivationFunctionProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.ActivationFunctionProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=186,
  serialized_end=265,
)


_LAYERPROTO = _descriptor.Descriptor(
  name='LayerProto',
  full_name='aihwx.LayerProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.LayerProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.LayerProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state_dict', full_name='aihwx.LayerProto.state_dict', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=267,
  serialized_end=376,
)


_LAYERORACTIVATIONFUNCTION = _descriptor.Descriptor(
  name='LayerOrActivationFunction',
  full_name='aihwx.LayerOrActivationFunction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='activation_function', full_name='aihwx.LayerOrActivationFunction.activation_function', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='layer', full_name='aihwx.LayerOrActivationFunction.layer', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='item', full_name='aihwx.LayerOrActivationFunction.item',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=379,
  serialized_end=513,
)


_OPTIMIZERPROTO = _descriptor.Descriptor(
  name='OptimizerProto',
  full_name='aihwx.OptimizerProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.OptimizerProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.OptimizerProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=515,
  serialized_end=585,
)


_SCHEDULERPROTO = _descriptor.Descriptor(
  name='SchedulerProto',
  full_name='aihwx.SchedulerProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.SchedulerProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.SchedulerProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=587,
  serialized_end=657,
)


_LOSSFUNCTIONPROTO = _descriptor.Descriptor(
  name='LossFunctionProto',
  full_name='aihwx.LossFunctionProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.LossFunctionProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.LossFunctionProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=659,
  serialized_end=732,
)


_NETWORK = _descriptor.Descriptor(
  name='Network',
  full_name='aihwx.Network',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='layers', full_name='aihwx.Network.layers', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=734,
  serialized_end=793,
)


_VERSION = _descriptor.Descriptor(
  name='Version',
  full_name='aihwx.Version',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schema', full_name='aihwx.Version.schema', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='opset', full_name='aihwx.Version.opset', index=1,
      number=2, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=795,
  serialized_end=835,
)

_PRIMITIVEPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_TRANSFORMPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_ACTIVATIONFUNCTIONPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_LAYERPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_LAYERPROTO.fields_by_name['state_dict'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_LAYERORACTIVATIONFUNCTION.fields_by_name['activation_function'].message_type = _ACTIVATIONFUNCTIONPROTO
_LAYERORACTIVATIONFUNCTION.fields_by_name['layer'].message_type = _LAYERPROTO
_LAYERORACTIVATIONFUNCTION.oneofs_by_name['item'].fields.append(
  _LAYERORACTIVATIONFUNCTION.fields_by_name['activation_function'])
_LAYERORACTIVATIONFUNCTION.fields_by_name['activation_function'].containing_oneof = _LAYERORACTIVATIONFUNCTION.oneofs_by_name['item']
_LAYERORACTIVATIONFUNCTION.oneofs_by_name['item'].fields.append(
  _LAYERORACTIVATIONFUNCTION.fields_by_name['layer'])
_LAYERORACTIVATIONFUNCTION.fields_by_name['layer'].containing_oneof = _LAYERORACTIVATIONFUNCTION.oneofs_by_name['item']
_OPTIMIZERPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_SCHEDULERPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_LOSSFUNCTIONPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_NETWORK.fields_by_name['layers'].message_type = _LAYERORACTIVATIONFUNCTION
DESCRIPTOR.message_types_by_name['PrimitiveProto'] = _PRIMITIVEPROTO
DESCRIPTOR.message_types_by_name['TransformProto'] = _TRANSFORMPROTO
DESCRIPTOR.message_types_by_name['ActivationFunctionProto'] = _ACTIVATIONFUNCTIONPROTO
DESCRIPTOR.message_types_by_name['LayerProto'] = _LAYERPROTO
DESCRIPTOR.message_types_by_name['LayerOrActivationFunction'] = _LAYERORACTIVATIONFUNCTION
DESCRIPTOR.message_types_by_name['OptimizerProto'] = _OPTIMIZERPROTO
DESCRIPTOR.message_types_by_name['SchedulerProto'] = _SCHEDULERPROTO
DESCRIPTOR.message_types_by_name['LossFunctionProto'] = _LOSSFUNCTIONPROTO
DESCRIPTOR.message_types_by_name['Network'] = _NETWORK
DESCRIPTOR.message_types_by_name['Version'] = _VERSION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PrimitiveProto = _reflection.GeneratedProtocolMessageType('PrimitiveProto', (_message.Message,), {
  'DESCRIPTOR' : _PRIMITIVEPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.PrimitiveProto)
  })
_sym_db.RegisterMessage(PrimitiveProto)

TransformProto = _reflection.GeneratedProtocolMessageType('TransformProto', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFORMPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.TransformProto)
  })
_sym_db.RegisterMessage(TransformProto)

ActivationFunctionProto = _reflection.GeneratedProtocolMessageType('ActivationFunctionProto', (_message.Message,), {
  'DESCRIPTOR' : _ACTIVATIONFUNCTIONPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.ActivationFunctionProto)
  })
_sym_db.RegisterMessage(ActivationFunctionProto)

LayerProto = _reflection.GeneratedProtocolMessageType('LayerProto', (_message.Message,), {
  'DESCRIPTOR' : _LAYERPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.LayerProto)
  })
_sym_db.RegisterMessage(LayerProto)

LayerOrActivationFunction = _reflection.GeneratedProtocolMessageType('LayerOrActivationFunction', (_message.Message,), {
  'DESCRIPTOR' : _LAYERORACTIVATIONFUNCTION,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.LayerOrActivationFunction)
  })
_sym_db.RegisterMessage(LayerOrActivationFunction)

OptimizerProto = _reflection.GeneratedProtocolMessageType('OptimizerProto', (_message.Message,), {
  'DESCRIPTOR' : _OPTIMIZERPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.OptimizerProto)
  })
_sym_db.RegisterMessage(OptimizerProto)

SchedulerProto = _reflection.GeneratedProtocolMessageType('SchedulerProto', (_message.Message,), {
  'DESCRIPTOR' : _SCHEDULERPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.SchedulerProto)
  })
_sym_db.RegisterMessage(SchedulerProto)

LossFunctionProto = _reflection.GeneratedProtocolMessageType('LossFunctionProto', (_message.Message,), {
  'DESCRIPTOR' : _LOSSFUNCTIONPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.LossFunctionProto)
  })
_sym_db.RegisterMessage(LossFunctionProto)

Network = _reflection.GeneratedProtocolMessageType('Network', (_message.Message,), {
  'DESCRIPTOR' : _NETWORK,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.Network)
  })
_sym_db.RegisterMessage(Network)

Version = _reflection.GeneratedProtocolMessageType('Version', (_message.Message,), {
  'DESCRIPTOR' : _VERSION,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.Version)
  })
_sym_db.RegisterMessage(Version)


# @@protoc_insertion_point(module_scope)
